/*
 Modified by Cameron Skinner
 May 2005: Converted to librescue
*/

// System.hxx
//
/////////////////////////////////////////////////////////////////////////////

#ifndef SYSTEM_H
#define SYSTEM_H

#include "../librescue/connection.h"
#include "../librescue/connection_manager.h"
#include "../librescue/config.h"
#include "../librescue/objectpool.h"
#include "../librescue/udp.h"
#include "../librescue/tcp.h"
#include "../librescue/input.h"
#include "../librescue/output.h"

#include <map>
#include "AutoPtr.hxx"

using namespace Librescue;

namespace Rescue
{
  /////////////////////////////////////////////////////////////////////////
  // System
    
  class Agent;
  /** kernel Υ */
  class System
  {
  private:
        
  protected:
	/** Хʥե */
	FILE* m_logFile;

	/** бѥå */
	ConnectionManager m_connectionManager;
	//	LongUDPConnection* m_publicSocket;
	/** ȱѥå */
	//	LongUDPConnection* m_agentSocket;
	/** ߥ졼ѥå */
	//	LongUDPConnection* m_simulatorSocket;
	/** GISѥå */
	//	LongUDPConnection* m_gisSocket;
	/**  */
	Config m_config;
	/** ߥ졼θ߻(ñ:ߥ졼󥵥) */
	S32 m_time;
	/** kernel θ߻(ñms) */
	long m_systemTime;
	/** ƥब٤ƤΥ֥Ȥν */
	ObjectPool m_objectPool;
	/** Хåե */
	Input m_inputBuffer;
	/** Хåե */
	Output m_outputBuffer;
	/** ɥ쥹Υꥹ */
	typedef std::list<Address> Addresses;
  public:
	/** Ack Ԥ֥ */
	class AckWaiter {
	public:
	  virtual ~AckWaiter() { }
	  virtual void send() = 0;    // ack Ԥåtime ιϸƤӤ¦Ԥȡ
	};
  protected:
	/** simulator Υɥ쥹 Ack Ԥ֥ȤؤΥޥå */
	typedef std::map<Address, AutoPtr<AckWaiter> > AddressToAckWaiter;
	AddressToAckWaiter m_addressToAckWaiter;
	/** agent  ID  Ack Ԥ֥ȤؤΥޥå */
	typedef std::map<Id, AutoPtr<AckWaiter> > IdToAckWaiter;
	IdToAckWaiter m_idToAckWaiter;

	/** ³Τ륷ߥ졼Υɥ쥹Υꥹ */
	Addresses m_simulators;
	/** ³ƤGISΥɥ쥹 */
	Address m_gis;
	/** agent ֥Ȥ TypeID Υꥹ */
	std::vector<TypeId> m_agentTypes;
	/** Agent ֥ȤΥꥹȷ */
	typedef std::list<Agent*> Agents;
	/** ³Τʤ Agent ֥ȤΥꥹ */
	Agents* m_pFreeAgents;
	/** ³Τ Agent ֥ȤΥꥹ */
	Agents* m_pConnectedAgents;
	/** ߼¹Ԥ褦ȤƤ륳ޥɤμ̤ Agent Υꥹ */
	typedef std::map<Header, Agents> HeaderToAgents;
	HeaderToAgents m_commandToAngents;
        
	/** Ȥ³Ԥ */
	bool startOk(); // ⤦ߥ졼򳫻ϤƤ⤤
	void setUp();
	bool m_gisConnected;    // GIS ³Ԥ³ʤʤ false³Ѥߤʤ true
	/** ̤³Ȥ򥷥ƥϿ */
	void addFreeAgent(Agent* agent);
	/** ᥤ롼פ¹Ԥ */
	virtual void loop();
	void loopCore(long& count, int division);
	virtual void resetAgents();
	virtual void resetSimulators();

	/** from åꡢ롣*/
	void receiveMessages(int timeout/*LongUDPConnection& from*/);
	/** å1Ľ롣³Ԥ٤ɤ֤ */
	bool processMessage(const Address& from/*const LongUDPConnection& from*/);
	virtual void processCommand(Agent* agent, Header header, S32 size, Input& input/*, const LongUDPConnection& from*/);
	void gisConnectOk(const Address& from/*const LongUDPConnection& from*/);
	void agentConnect(const Address& from/*const LongUDPConnection& from*/);
	void agentAcknowledge(/*const LongUDPConnection& from*/);
	void agentConnectError(const Address& from,/*const LongUDPConnection& from,*/ S32 temporaryId, const char* message);
	void simulatorConnect(const Address& from/*const LongUDPConnection& from*/);
	void simulatorAcknowledge(const Address& from/*const LongUDPConnection& from*/);
        
	/** Ȥؾ */
	virtual void sendToAgents() = 0;
	/** ߥ졼ؾ */
	void sendToSimulators();
	/** ߥ졼GISع */
	void sendUpdate(S32 time);
	/** ѽ */
	void step();

	// Find the Agent that controls a particular RescueObject
	Agent* findAgent(RescueObject* object);

  public:
	virtual void outputObjectForAgentAtStart(Agent* agent, const RescueObject* o, Output& buffer);
	virtual void outputObjectForAgent(Agent* agent, const RescueObject* o, Output& buffer);

	virtual ~System();
	System(const std::vector<TypeId>& agentTypes);

	/** бѥå */
	//	LongUDPConnection& publicSocket();
	/** ȱѥå */
	//	LongUDPConnection& agentSocket();
	/** ߥ졼ѥå */
	//	LongUDPConnection& simulatorSocket();
	/** GISѥå */
	//	LongUDPConnection& gisSocket();
	/**  */
	Config& config();
	/** ƥब٤ƤΥ֥Ȥν */
	ObjectPool& objectPool();

	/* ::main Ϥ줿ѥ᥿ǥƥ¹Ԥ */
	void main(int argc, char const* const* argv); //throw(std::exception)
        
  private:
	System(const System& source);
	System& operator= (const System& rhs);
	bool operator== (const System& rhs) const;
  };
    
  /////////////////////////////////////////////////////////////////////////
} // namespace Rescue

#endif // } !defined(SYSTEM_HXX__INCLUDED)
